import math
import sys
import requests
from PIL import Image
import io

# コマンドライン引数の処理：タイルの種類、ズームレベル、地理的範囲（西、東、南、北の度数）、
# 出力ファイル名を引数として受け取ります。
# 地理的座標からタイル座標への変換：緯度経度をタイルの座標に変換します。
# タイル画像の取得と合成：指定された範囲のタイルをダウンロードし、一つの画像に合成します。
# 画像の保存：合成された画像をPNG形式で保存します。

# requestsはHTTPリクエスト用、Pillowは画像処理 module
# conda install pillow


# tile_type: std,

# 特定のパラメータ（タイルのタイプ、ズームレベル、x座標、y座標）を受け取り、
# それを元に地理情報システム（GIS）のタイルに対するURLを生成する関数
def gsitile_url(tile_type, zoom, x, y):
    base_url = "https://cyberjapandata.gsi.go.jp/xyz/"
    url = f"{base_url}{tile_type}/{zoom}/{x}/{y}.png"
    return url


def deg2rad(deg):
    return deg * math.pi / 180.0

def rad2deg(rad):
    return rad * 180.0 / math.pi

def xy2ll(x, y):
    if abs(x) > math.pi or abs(y) > math.pi:
        return None
    lon_deg = rad2deg(x)
    lon_deg = -180.0 if lon_deg < -180.0 else lon_deg
    lon_deg =  180.0 if lon_deg >  180.0 else lon_deg
    lat_deg = rad2deg(math.asin(math.tanh(y)))
    return lon_deg, lat_deg

def ll2xy(lon_deg, lat_deg):
    if lon_deg > 180.0:
        lon_deg -= 360.0
    if abs(lon_deg) > 180.0 or abs(lat_deg) > 90.0:
        return None
    x = deg2rad(lon_deg)
    x = -math.pi if x < -math.pi else x
    x =  math.pi if x >  math.pi else x
    y = math.atanh(math.sin(deg2rad(lat_deg)))
    if abs(y) > math.pi:
        return None
    return x, y

def xy2ll_hash(x, y):
    ll = xy2ll(x, y)
    if ll is None:
        return None
    return {'lon': ll[0], 'lat': ll[1]}

def ll2xy_hash(lon_deg, lat_deg):
    result = ll2xy(lon_deg, lat_deg)
    if result is None:
        return None
    return {'x': result[0], 'y': result[1]}

def ztile2ll(z, tile):
    if not all(k in tile for k in ('ti', 'tj', 'i', 'j')):
        return None

    t = 2.0 * math.pi / (2 ** z)
    xx = t * (tile['ti'] + (tile['i'] + 0.5) / 256.0)
    yy = t * (tile['tj'] + (tile['j'] + 0.5) / 256.0)
    x = xx - math.pi
    y = math.pi - yy
    return xy2ll_hash(x, y)

def zll2tile(z, lon, lat):
    xy = ll2xy_hash(lon, lat)
    if xy is None:
        return None

    xx = math.pi + xy['x']
    yy = math.pi - xy['y']
    t = 2.0 * math.pi / (2 ** z)
    ti = int(xx / t)
    tj = int(yy / t)
    i = int((xx - ti * t) / t * 256)
    i = 255 if i > 255 else i
    j = int((yy - tj * t) / t * 256)
    j = 255 if j > 255 else j
    return {'ti': ti, 'tj': tj, 'i': i, 'j': j}

def calculate_dimensions(tile_nw, tile_se):
    if tile_nw['ti'] == tile_se['ti']:
        width = tile_se['i'] - tile_nw['i'] + 1
    else:
        width = (256 - tile_nw['i']) + (tile_se['i'] + 1) \
            + (tile_se['ti'] - tile_nw['ti'] - 1) * 256

    if tile_nw['tj'] == tile_se['tj']:
        height = tile_se['j'] - tile_nw['j'] + 1
    else:
        height = (256 - tile_nw['j']) + (tile_se['j'] + 1) \
            + (tile_se['tj'] - tile_nw['tj'] - 1) * 256

    return width, height

def gsitile2png(tile_type, zoom, west, east, south, north, out):
    
    zoom = int(zoom)
    tile_nw = zll2tile(zoom, float(west), float(north))
    tile_se = zll2tile(zoom, float(east), float(south))
    
    if tile_nw is None or tile_se is None:
        print("Bad geometries", file=sys.stderr)
        sys.exit(1)

    width, height = calculate_dimensions(tile_nw, tile_se)
    
    if out == '--size':
        print(f"{width} {height}")
        sys.exit(0)

    # HTTPリクエスト用のセッションを作成し、タイムアウトを設定
    session = requests.Session()
    session.timeout = 10

    # 新しい画像オブジェクトの作成
    img = Image.new('RGB', (width, height))

    for tj in range(tile_nw['tj'], tile_se['tj'] + 1):
        for ti in range(tile_nw['ti'], tile_se['ti'] + 1):

            # 大きな地図画像の中で各タイルがどの位置に配置されるかを決定
            # dstx は目的の画像（大きな地図画像）の中でのタイルのx座標、
            # srcx はソースタイル画像のx座標、w はソースタイルからコピーする
            # 幅を表します。この計算は、組み合わせる地図タイルが左端、右端、
            # または中間にあるかによって異なります。
            if ti == tile_nw['ti']:
                dstx = 0
                srcx = tile_nw['i']
                w = 256 - tile_nw['i']
            elif ti == tile_se['ti']:
                dstx = width - (tile_se['i'] + 1)
                srcx = 0
                w = tile_se['i'] + 1
            else:
                dstx = 256 - tile_nw['i'] + 256 * (ti - tile_nw['ti'] - 1)
                srcx = 0
                w = 256

            if tj == tile_nw['tj']:
                dsty = 0
                srcy = tile_nw['j']
                h = 256 - tile_nw['j']
            elif tj == tile_se['tj']:
                dsty = height - (tile_se['j'] + 1)
                srcy = 0
                h = tile_se['j'] + 1
            else:
                dsty = 256 - tile_nw['j'] + 256 * (tj - tile_nw['tj'] - 1)
                srcy = 0
                h = 256

            url = gsitile_url(tile_type, zoom, ti, tj)
            response = session.get(url)

            if not response.ok:
                print(f"Cannot get {url} ({response.status_code})", \
                      file=sys.stderr)
                sys.exit(1)

            # レスポンスのコンテンツ（バイナリデータ）からPillow画像オブジェクトを作成
            try:
                timg = Image.open(io.BytesIO(response.content))
            except IOError:
                print("Cannot create tile image", file=sys.stderr)
                sys.exit(1)

            timg_cropped = timg.crop((srcx, srcy, srcx + w, srcy + h))
            img.paste(timg_cropped, (dstx, dsty))   

    # 画像をPNGフォーマットでファイルに保存
    img.save(out, format='PNG')            

